@@ -35,6 +35,7 @@
 
 #include <klocalizedstring.h>
 #include <kdeclarative/kdeclarative.h>
+#include <KQuickAddons/ConfigModule>
 
 #include <Plasma/Corona>
 #include <Plasma/PluginLoader>
@@ -53,6 +54,7 @@ public:
     ConfigModel *q;
     QList<ConfigCategory *> categories;
     QWeakPointer<Plasma::Applet> appletInterface;
+    QHash<QString, KQuickAddons::ConfigModule *> kcms;
 
     void appendCategory(ConfigCategory *c);
     void clear();
@@ -144,15 +146,16 @@ QVariant ConfigModelPrivate::get(int row) const
         return value;
     }
 
-    value["name"] = categories.at(row)->name();
-    value["icon"] = categories.at(row)->icon();
-    value["pluginName"] = categories.at(row)->pluginName();
+    value[QStringLiteral("name")] = categories.at(row)->name();
+    value[QStringLiteral("icon")] = categories.at(row)->icon();
+    value[QStringLiteral("pluginName")] = categories.at(row)->pluginName();
     if (appletInterface) {
-        value["source"] = QUrl::fromLocalFile(appletInterface.data()->package().filePath("ui", categories.at(row)->source()));
+        value[QStringLiteral("source")] = QUrl::fromLocalFile(appletInterface.data()->package().filePath("ui", categories.at(row)->source()));
     } else {
-        value["source"] = categories.at(row)->source();
+        value[QStringLiteral("source")] = categories.at(row)->source();
     }
-    value["visible"] = categories.at(row)->visible();
+    value[QStringLiteral("visible")] = categories.at(row)->visible();
+    value[QStringLiteral("kcm")] = q->data(q->index(row, 0), ConfigModel::KCMRole);
 
     return value;
 }
@@ -167,6 +170,7 @@ ConfigModel::ConfigModel(QObject *parent)
     roleNames[SourceRole] = "source";
     roleNames[PluginNameRole] = "pluginName";
     roleNames[VisibleRole] = "visible";
+    roleNames[KCMRole] = "kcm";
 
     setRoleNames(roleNames);
 }
@@ -195,15 +199,44 @@ QVariant ConfigModel::data(const QModelIndex &index, int role) const
     case IconRole:
         return d->categories.at(index.row())->icon();
     case SourceRole:
-        if (d->appletInterface) {
-            return QUrl::fromLocalFile(d->appletInterface.data()->package().filePath("ui", d->categories.at(index.row())->source()));
+    {
+        const QString source = d->categories.at(index.row())->source();
+        // Quick check if source is an absolute path or not
+        if (d->appletInterface && !(source.startsWith('/') && source.endsWith(QLatin1String("qml")))) {
+            return QUrl::fromLocalFile(d->appletInterface.data()->package().filePath("ui", source));
         } else {
-            return d->categories.at(index.row())->source();
+            return source;
         }
+    }
     case PluginNameRole:
         return d->categories.at(index.row())->pluginName();
     case VisibleRole:
         return d->categories.at(index.row())->visible();
+    case KCMRole: {
+        const QString pluginName = d->categories.at(index.row())->pluginName();
+        const QString pluginPath = KPluginLoader::findPlugin(pluginName);
+        //no kcm is registered for this row, it's a normal qml-only entry
+        if (pluginName.isEmpty() || pluginPath.isEmpty()) {
+            return QVariant();
+        }
+
+        if (d->kcms.contains(pluginName)) {
+            return QVariant::fromValue(d->kcms.value(pluginName));
+        }
+
+        KPluginLoader loader(pluginPath);
+        KPluginFactory* factory = loader.factory();
+        if (!factory) {
+            qWarning() << "Error loading KCM:" << loader.errorString();
+        } else {
+            KQuickAddons::ConfigModule *cm = factory->create<KQuickAddons::ConfigModule >(const_cast<ConfigModel *>(this));
+            if (!cm) {
+                qWarning() << "Error creating KCM object from plugin" << loader.fileName();
+            }
+            d->kcms[pluginName] = cm;
+            return QVariant::fromValue(cm);
+        }
+    }
     default:
         return QVariant();
     }
