@@ -62,6 +62,7 @@ public:
         : q(dialog),
           location(Plasma::Types::BottomEdge),
           frameSvgItem(0),
+          hasMask(false),
           type(Dialog::Normal),
           hideOnWindowDeactivate(false),
           outputOnly(false),
@@ -131,6 +132,7 @@ public:
     QTimer hintsCommitTimer;
 
     QRect cachedGeometry;
+    bool hasMask;
     Dialog::WindowType type;
     bool hideOnWindowDeactivate;
     bool outputOnly;
@@ -152,7 +154,7 @@ QRect DialogPrivate::availableScreenGeometryForPosition(const QPoint& pos) const
     //        we simply iterate over the virtual screens and pick the one our QWindow
     //        says it's at.
     QRect avail;
-    Q_FOREACH (QScreen *screen, q->screen()->virtualSiblings()) {
+    Q_FOREACH (QScreen *screen, QGuiApplication::screens()) {
         //we check geometry() but then take availableGeometry()
         //to reliably check in which screen a position is, we need the full
         //geometry, including areas for panels
@@ -168,8 +170,10 @@ QRect DialogPrivate::availableScreenGeometryForPosition(const QPoint& pos) const
      * the screen should be correctly updated now on Qt 5.3+ so should be
      * more reliable anyways (could be tried to remove the whole Q_FOREACH
      * at this point)
+     *
+     * imporant: screen can be a nullptr... see bug 345173
      */
-    if (avail.isEmpty()) {
+    if (avail.isEmpty() && q->screen()) {
         avail = q->screen()->availableGeometry();
     }
 
@@ -213,9 +217,9 @@ void DialogPrivate::updateTheme()
         DialogShadows::self()->removeWindow(q);
     } else {
         if (type == Dialog::Tooltip) {
-            frameSvgItem->setImagePath("widgets/tooltip");
+            frameSvgItem->setImagePath(QStringLiteral("widgets/tooltip"));
         } else {
-            frameSvgItem->setImagePath("dialogs/background");
+            frameSvgItem->setImagePath(QStringLiteral("dialogs/background"));
         }
 
         KWindowEffects::enableBlurBehind(q->winId(), true, frameSvgItem->frameSvg()->mask());
@@ -227,8 +231,12 @@ void DialogPrivate::updateTheme()
                 frameSvgItem->frameSvg()->mask());
 
         if (KWindowSystem::compositingActive()) {
-            q->setMask(QRegion());
+            if (hasMask) {
+                hasMask = false;
+                q->setMask(QRegion());
+            }
         } else {
+            hasMask = true;
             q->setMask(frameSvgItem->frameSvg()->mask());
         }
         if (q->isVisible()) {
@@ -277,7 +285,7 @@ void DialogPrivate::updateVisibility(bool visible)
     }
 
 
-    if (!(q->flags() & Qt::ToolTip)) {
+    if (!(q->flags() & Qt::ToolTip) && type != Dialog::Notification) {
         KWindowEffects::SlideFromLocation slideLocation = KWindowEffects::NoEdge;
 
         switch (location) {
@@ -659,7 +667,7 @@ Dialog::Dialog(QQuickItem *parent)
     setColor(QColor(Qt::transparent));
     setFlags(Qt::FramelessWindowHint);
 
-    setIcon(QIcon::fromTheme("plasma"));
+    setIcon(QIcon::fromTheme(QStringLiteral("plasma")));
 
     connect(this, &QWindow::xChanged, [=]() { d->slotWindowPositionChanged(); });
     connect(this, &QWindow::yChanged, [=]() { d->slotWindowPositionChanged(); });
@@ -996,9 +1004,9 @@ void Dialog::setType(WindowType type)
         d->frameSvgItem->setImagePath(QString());
     } else {
         if (d->type == Tooltip) {
-            d->frameSvgItem->setImagePath("widgets/tooltip");
+            d->frameSvgItem->setImagePath(QStringLiteral("widgets/tooltip"));
         } else {
-            d->frameSvgItem->setImagePath("dialogs/background");
+            d->frameSvgItem->setImagePath(QStringLiteral("dialogs/background"));
         }
     }
 
@@ -1063,7 +1071,9 @@ void Dialog::showEvent(QShowEvent *event)
 
 bool Dialog::event(QEvent *event)
 {
-    if (event->type() == QEvent::Show) {
+    if (event->type() == QEvent::Expose) {
+        KWindowSystem::setState(winId(), NET::SkipTaskbar | NET::SkipPager);
+    } else if (event->type() == QEvent::Show) {
         d->updateVisibility(true);
     } else if (event->type() == QEvent::Hide) {
         d->updateVisibility(false);
@@ -1161,11 +1171,7 @@ bool Dialog::event(QEvent *event)
         }
     }
 
-    const bool retval = QQuickWindow::event(event);
-    if (event->type() != QEvent::DeferredDelete) {
-        KWindowSystem::setState(winId(), NET::SkipTaskbar | NET::SkipPager);
-    }
-    return retval;
+    return QQuickWindow::event(event);
 }
 
 void Dialog::hideEvent(QHideEvent *event)
@@ -1229,7 +1235,7 @@ void Dialog::setVisible(bool visible)
 
     d->visible = visible;
     if (d->componentComplete) {
-        if (d->visualParent) {
+        if (visible && d->visualParent) {
             setPosition(popupPosition(d->visualParent, size()));
         }
         QQuickWindow::setVisible(visible);
