@@ -34,6 +34,7 @@
 #include <klocalizedstring.h>
 #include <kurlmimedata.h>
 #include <QMimeDatabase>
+#include <KNotification>
 
 #ifndef PLASMA_NO_KIO
 #include "kio/jobclasses.h" // for KIO::JobFlags
@@ -44,9 +45,11 @@
 #include <plasma.h>
 #include <Plasma/ContainmentActions>
 #include <Plasma/Corona>
-#include <Plasma/Package>
 #include <Plasma/PluginLoader>
 
+#include <KPackage/Package>
+#include <KPackage/PackageLoader>
+
 #include <kactivities/info.h>
 
 #include "kdeclarative/configpropertymap.h"
@@ -55,7 +58,8 @@
 ContainmentInterface::ContainmentInterface(DeclarativeAppletScript *parent, const QVariantList &args)
     : AppletInterface(parent, args),
       m_wallpaperInterface(0),
-      m_activityInfo(0)
+      m_activityInfo(0),
+      m_wheelDelta(0)
 {
     m_containment = static_cast<Plasma::Containment *>(appletScript()->applet()->containment());
 
@@ -71,13 +75,16 @@ ContainmentInterface::ContainmentInterface(DeclarativeAppletScript *parent, cons
     }
 
     connect(QCoreApplication::instance(), &QCoreApplication::aboutToQuit,
-            [=]() {
-                if (!m_containment) {
-                    return;
-                }
-                disconnect(m_containment.data(), &Plasma::Containment::appletRemoved,
-                           this, &ContainmentInterface::appletRemovedForward);
-            });
+            this, &ContainmentInterface::slotAboutToQuit);
+}
+
+void ContainmentInterface::slotAboutToQuit()
+{
+    if (!m_containment) {
+        return;
+    }
+    disconnect(m_containment.data(), &Plasma::Containment::appletRemoved,
+               this, &ContainmentInterface::appletRemovedForward);
 }
 
 void ContainmentInterface::init()
@@ -102,31 +109,39 @@ void ContainmentInterface::init()
             defaults = KConfigGroup(KSharedConfig::openConfig(m_containment->corona()->package().filePath("defaults")), "Panel");
         }
 
-        Plasma::Package pkg = Plasma::PluginLoader::self()->loadPackage("Plasma/Generic");
-
         if (defaults.isValid()) {
-            pkg.setPath(defaults.readEntry("ToolBox", "org.kde.desktoptoolbox"));
-        } else {
-            pkg.setPath("org.kde.desktoptoolbox");
-        }
+            KPackage::Package pkg = KPackage::PackageLoader::self()->loadPackage("KPackage/GenericQML");
+            pkg.setDefaultPackageRoot("plasma/packages");
 
-        PlasmaQuick::PackageUrlInterceptor *interceptor = dynamic_cast<PlasmaQuick::PackageUrlInterceptor *>(qmlObject()->engine()->urlInterceptor());
-        if (interceptor) {
-            interceptor->addAllowedPath(pkg.path());
-        }
+            if (defaults.isValid()) {
+                pkg.setPath(defaults.readEntry("ToolBox", "org.kde.desktoptoolbox"));
+            } else {
+                pkg.setPath("org.kde.desktoptoolbox");
+            }
 
-        if (pkg.isValid()) {
-            QObject *containmentGraphicObject = qmlObject()->rootObject();
+            PlasmaQuick::PackageUrlInterceptor *interceptor = dynamic_cast<PlasmaQuick::PackageUrlInterceptor *>(qmlObject()->engine()->urlInterceptor());
+            if (interceptor) {
+                interceptor->addAllowedPath(pkg.path());
+            }
 
-            QVariantHash toolboxProperties;
-            toolboxProperties["parent"] = QVariant::fromValue(this);
-            QObject *toolBoxObject = qmlObject()->createObjectFromSource(QUrl::fromLocalFile(pkg.filePath("mainscript")), 0, toolboxProperties);
-            if (toolBoxObject && containmentGraphicObject) {
-                containmentGraphicObject->setProperty("toolBox", QVariant::fromValue(toolBoxObject));
+            if (pkg.metadata().isValid() && !pkg.metadata().isHidden()) {
+                if (pkg.isValid()) {
+                    QObject *containmentGraphicObject = qmlObject()->rootObject();
+
+                    QVariantHash toolboxProperties;
+                    toolboxProperties["parent"] = QVariant::fromValue(this);
+                    QObject *toolBoxObject = qmlObject()->createObjectFromSource(QUrl::fromLocalFile(pkg.filePath("mainscript")), 0, toolboxProperties);
+                    if (toolBoxObject && containmentGraphicObject) {
+                        containmentGraphicObject->setProperty("toolBox", QVariant::fromValue(toolBoxObject));
+                    }
+                } else {
+                    qWarning() << "Could not load toolbox package." << pkg.path();
+                }
+            } else {
+                qWarning() << "Toolbox not loading, toolbox package is either invalid or disabled.";
             }
-        } else {
-            qWarning() << "Could not load toolbox package." << pkg.path();
         }
+
     }
 
     //set parent, both as object hierarchically and visually
@@ -135,8 +150,8 @@ void ContainmentInterface::init()
         qmlObject()->rootObject()->setProperty("parent", QVariant::fromValue(this));
 
         //set anchors
-        QQmlExpression expr(qmlObject()->engine()->rootContext(), qmlObject()->rootObject(), "parent");
-        QQmlProperty prop(qmlObject()->rootObject(), "anchors.fill");
+        QQmlExpression expr(qmlObject()->engine()->rootContext(), qmlObject()->rootObject(), QStringLiteral("parent"));
+        QQmlProperty prop(qmlObject()->rootObject(), QStringLiteral("anchors.fill"));
         prop.write(expr.evaluate());
     }
 
@@ -159,6 +174,11 @@ void ContainmentInterface::init()
     connect(m_containment.data(), &Plasma::Containment::containmentTypeChanged,
             this, &ContainmentInterface::containmentTypeChanged);
 
+    connect(m_containment.data()->actions(), &KActionCollection::inserted,
+            this, &ContainmentInterface::actionsChanged);
+    connect(m_containment.data()->actions(), &KActionCollection::removed,
+            this, &ContainmentInterface::actionsChanged);
+
     if (m_containment->corona()) {
         connect(m_containment->corona(), &Plasma::Corona::availableScreenRegionChanged,
                 this, &ContainmentInterface::availableScreenRegionChanged);
@@ -250,17 +270,22 @@ void ContainmentInterface::setAppletArgs(Plasma::Applet *applet, const QString &
 
 QObject *ContainmentInterface::containmentAt(int x, int y)
 {
+    QObject *desktop = Q_NULLPTR;
     foreach (Plasma::Containment *c, m_containment->corona()->containments()) {
         ContainmentInterface *contInterface = c->property("_plasma_graphicObject").value<ContainmentInterface *>();
 
         if (contInterface && contInterface->isVisible()) {
             QWindow *w = contInterface->window();
             if (w && w->geometry().contains(QPoint(window()->x(), window()->y()) + QPoint(x, y))) {
-                return contInterface;
+                if (c->containmentType() == Plasma::Types::DesktopContainment) {
+                    desktop = contInterface;
+                } else {
+                    return contInterface;
+                }
             }
         }
     }
-    return 0;
+    return desktop;
 }
 
 void ContainmentInterface::addApplet(AppletInterface *applet, int x, int y)
@@ -303,6 +328,125 @@ QPointF ContainmentInterface::mapToApplet(AppletInterface *applet, int x, int y)
     return pos - applet->mapToScene(QPointF(0, 0));
 }
 
+QPointF ContainmentInterface::adjustToAvailableScreenRegion(int x, int y, int w, int h) const
+{
+    QRegion reg;
+    int screenId = screen();
+    if (screenId > -1 && m_containment->corona()) {
+        reg = m_containment->corona()->availableScreenRegion(screenId);
+    }
+
+    if (!reg.isEmpty()) {
+        //make it relative
+        QRect geometry = m_containment->corona()->screenGeometry(screenId);
+        reg.translate(- geometry.topLeft());
+    } else {
+        reg = QRect(0, 0, width(), height());
+    }
+
+    const QRect rect(qBound(reg.boundingRect().left(), x, reg.boundingRect().right() + 1 - w),
+                     qBound(reg.boundingRect().top(), y, reg.boundingRect().bottom() + 1 - h), w, h);
+    const QRectF ar = availableScreenRect();
+    QRect tempRect(rect);
+
+    // in the case we are in the topleft quadrant
+    // * see if the passed rect is completely in the region, if yes, return
+    // * otherwise, try to move it horizontally to the screenrect x
+    // * if now fits, return
+    // * if fail, move vertically
+    // * as last resort, move horizontally and vertically
+
+    // top left corner
+    if (rect.center().x() <= ar.center().x() && rect.center().y() <= ar.center().y()) {
+        //QRegion::contains doesn't do what it would suggest, so do reg.intersected(rect) != rect instead
+        if (reg.intersected(rect) != rect) {
+            tempRect = QRect(qMax(rect.left(), (int)ar.left()), rect.top(), w, h);
+            if (reg.intersected(tempRect) == tempRect) {
+                return tempRect.topLeft();
+            }
+
+            tempRect = QRect(rect.left(), qMax(rect.top(), (int)ar.top()), w, h);
+            if (reg.intersected(tempRect) == tempRect) {
+                return tempRect.topLeft();
+            }
+
+            tempRect = QRect(qMax(rect.left(), (int)ar.left()), qMax(rect.top(), (int)ar.top()), w, h);
+            return tempRect.topLeft();
+        } else {
+            return rect.topLeft();
+        }
+
+    //bottom left corner
+    } else if (rect.center().x() <= ar.center().x() && rect.center().y() > ar.center().y()) {
+        if (reg.intersected(rect) != rect) {
+            tempRect = QRect(qMax(rect.left(), (int)ar.left()), rect.top(), w, h);
+            if (reg.intersected(tempRect) == tempRect) {
+                return tempRect.topLeft();
+            }
+
+            tempRect = QRect(rect.left(), qMin(rect.top(), (int)(ar.bottom() + 1 - h)), w, h);
+            if (reg.intersected(tempRect) == tempRect) {
+                return tempRect.topLeft();
+            }
+
+            tempRect = QRect(qMax(rect.left(), (int)ar.left()), qMin(rect.top(), (int)(ar.bottom() + 1 - h)), w, h);
+            return tempRect.topLeft();
+        } else {
+            return rect.topLeft();
+        }
+
+    //top right corner
+    } else if (rect.center().x() > ar.center().x() && rect.center().y() <= ar.center().y()) {
+        if (reg.intersected(rect) != rect) {
+            tempRect = QRect(qMin(rect.left(), (int)(ar.right() + 1 - w)), rect.top(), w, h);
+            if (reg.intersected(tempRect) == tempRect) {
+                return tempRect.topLeft();
+            }
+
+            tempRect = QRect(rect.left(), qMax(rect.top(), (int)ar.top()), w, h);
+            if (reg.intersected(tempRect) == tempRect) {
+                return tempRect.topLeft();
+            }
+
+            tempRect = QRect(qMin(rect.left(), (int)(ar.right() + 1 - w)), qMax(rect.top(), (int)ar.top()), w, h);
+            return tempRect.topLeft();
+        } else {
+            return rect.topLeft();
+        }
+
+    //bottom right corner
+    } else if (rect.center().x() > ar.center().x() && rect.center().y() > ar.center().y()) {
+        if (reg.intersected(rect) != rect) {
+            tempRect = QRect(qMin(rect.left(), (int)(ar.right() + 1 - w)), rect.top(), w, h);
+            if (reg.intersected(tempRect) == tempRect) {
+                return tempRect.topLeft();
+            }
+
+            tempRect = QRect(rect.left(), qMin(rect.top(), (int)(ar.bottom() + 1 - h)), w, h);
+            if (reg.intersected(tempRect) == tempRect) {
+                return tempRect.topLeft();
+            }
+
+            tempRect = QRect(qMin(rect.left(), (int)(ar.right() + 1 - w)), qMin(rect.top(), (int)(ar.bottom() + 1 - h)), w, h);
+            return tempRect.topLeft();
+        } else {
+            return rect.topLeft();
+        }
+    }
+
+    return rect.topLeft();
+}
+
+void ContainmentInterface::processMimeData(QObject *mimeDataProxy, int x, int y)
+{
+    QMimeData* mime = qobject_cast<QMimeData*>(mimeDataProxy);
+    if (mime) {
+        processMimeData(mime, x, y);
+    } else {
+        processMimeData(mimeDataProxy->property("mimeData").value<QMimeData*>(), x, y);
+    }
+}
+
 void ContainmentInterface::processMimeData(QMimeData *mimeData, int x, int y)
 {
     if (!mimeData) {
@@ -313,8 +457,8 @@ void ContainmentInterface::processMimeData(QMimeData *mimeData, int x, int y)
 
     qDebug() << "Arrived mimeData" << mimeData->urls() << mimeData->formats() << "at" << x << ", " << y;
 
-    if (mimeData->hasFormat("text/x-plasmoidservicename")) {
-        QString data = mimeData->data("text/x-plasmoidservicename");
+    if (mimeData->hasFormat(QStringLiteral("text/x-plasmoidservicename"))) {
+        QString data = mimeData->data(QStringLiteral("text/x-plasmoidservicename"));
         const QStringList appletNames = data.split('\n', QString::SkipEmptyParts);
         foreach (const QString &appletName, appletNames) {
             qDebug() << "adding" << appletName;
@@ -344,8 +488,8 @@ void ContainmentInterface::processMimeData(QMimeData *mimeData, int x, int y)
             QObject::connect(job, SIGNAL(mimetype(KIO::Job*,QString)),
                              this, SLOT(mimeTypeRetrieved(KIO::Job*,QString)));
 
-            QMenu *choices = new QMenu("Content dropped");
-            choices->addAction(QIcon::fromTheme("process-working"), i18n("Fetching file type..."));
+            QMenu *choices = new QMenu(i18n("Content dropped"));
+            choices->addAction(QIcon::fromTheme(QStringLiteral("process-working")), i18n("Fetching file type..."));
             choices->popup(window() ? window()->mapToGlobal(QPoint(x, y)) : QPoint(x, y));
 
             m_dropMenus[job] = choices;
@@ -440,7 +584,7 @@ void ContainmentInterface::mimeTypeRetrieved(KIO::Job *job, const QString &mimet
         return;
     }
     KPluginInfo::List appletList = Plasma::PluginLoader::self()->listAppletInfoForUrl(tjob->url());
-    if (mimetype.isEmpty() && !appletList.count()) {
+    if (mimetype.isEmpty() && appletList.isEmpty()) {
         clearDataForMimeJob(job);
         qDebug() << "No applets found matching the url (" << tjob->url() << ") or the mimetype (" << mimetype << ")";
         return;
@@ -466,16 +610,28 @@ void ContainmentInterface::mimeTypeRetrieved(KIO::Job *job, const QString &mimet
         qDebug() << "Creating menu for:" << mimetype  << posi;
 
         appletList << Plasma::PluginLoader::self()->listAppletInfoForMimeType(mimetype);
+
         KPluginInfo::List wallpaperList;
 
-        if (m_wallpaperInterface && m_wallpaperInterface->supportsMimetype(mimetype)) {
-            wallpaperList << m_wallpaperInterface->package().metadata();
-        } else {
-            wallpaperList = WallpaperInterface::listWallpaperInfoForMimetype(mimetype);
+        if (m_containment->containmentType() != Plasma::Types::PanelContainment
+            && m_containment->containmentType() != Plasma::Types::CustomPanelContainment) {
+
+            if (m_wallpaperInterface && m_wallpaperInterface->supportsMimetype(mimetype)) {
+                wallpaperList << m_wallpaperInterface->package().metadata();
+            } else {
+                wallpaperList = WallpaperInterface::listWallpaperInfoForMimetype(mimetype);
+            }
         }
 
-        if (!appletList.isEmpty() || !wallpaperList.isEmpty()) {
-            choices->clear();
+        const bool isPlasmaPackage = (mimetype == QLatin1String("application/x-plasma"));
+
+        if (!appletList.isEmpty() || !wallpaperList.isEmpty() || isPlasmaPackage) {
+            QAction *installPlasmaPackageAction = nullptr;
+            if (isPlasmaPackage) {
+                choices->addSection(i18n("Plasma Package"));
+                installPlasmaPackageAction = choices->addAction(QIcon::fromTheme(QStringLiteral("application-x-plasma")), i18n("Install"));
+            }
+
             QHash<QAction *, QString> actionsToApplets;
             choices->addSection(i18n("Widgets"));
             foreach (const KPluginInfo &info, appletList) {
@@ -490,7 +646,7 @@ void ContainmentInterface::mimeTypeRetrieved(KIO::Job *job, const QString &mimet
                 actionsToApplets.insert(action, info.pluginName());
                 qDebug() << info.pluginName();
             }
-            actionsToApplets.insert(choices->addAction(i18n("Icon")), "org.kde.plasma.icon");
+            actionsToApplets.insert(choices->addAction(i18n("Icon")), QStringLiteral("org.kde.plasma.icon"));
 
             QHash<QAction *, QString> actionsToWallpapers;
             if (!wallpaperList.isEmpty())  {
@@ -513,6 +669,11 @@ void ContainmentInterface::mimeTypeRetrieved(KIO::Job *job, const QString &mimet
                 }
             }
 
+            // HACK If the QMenu becomes empty at any point after the "determining mimetype"
+            // popup was shown, it self-destructs, does not matter if we call clear() or remove
+            // the action manually, hence we remove the aforementioned item after we populated the menu
+            choices->removeAction(choices->actions().at(0));
+
             QAction *choice = choices->exec();
             if (choice) {
                 // Put the job on hold so it can be recycled to fetch the actual content,
@@ -522,8 +683,45 @@ void ContainmentInterface::mimeTypeRetrieved(KIO::Job *job, const QString &mimet
                     tjob->putOnHold();
                     KIO::Scheduler::publishSlaveOnHold();
                 }
+
                 QString plugin = actionsToApplets.value(choice);
-                if (plugin.isEmpty()) {
+
+                if (choice == installPlasmaPackageAction) {
+                    using namespace KPackage;
+                    PackageStructure *structure = PackageLoader::self()->loadPackageStructure(QStringLiteral("Plasma/Applet"));
+                    Package package(structure);
+
+                    const QString &packagePath = tjob->url().toLocalFile();
+
+                    KJob *installJob = package.update(packagePath);
+                    connect(installJob, &KJob::result, this, [this, packagePath, structure, posi](KJob *job) {
+                        auto fail = [job](const QString &text) {
+                            KNotification::event(QStringLiteral("plasmoidInstallationFailed"), i18n("Package Installation Failed"),
+                                                 text, QStringLiteral("dialog-error"), 0, KNotification::CloseOnTimeout, QStringLiteral("plasma_workspace"));
+                        };
+
+                        // if the applet is already installed, just add it to the containment
+                        if (job->error() != KJob::NoError
+                            && job->error() != Package::PackageAlreadyInstalledError
+                            && job->error() != Package::NewerVersionAlreadyInstalledError) {
+                            fail(job->errorText());
+                            return;
+                        }
+
+                        using namespace KPackage;
+                        Package package(structure);
+                        // TODO how can I get the path of the actual package?
+                        package.setPath(packagePath);
+
+                        // TODO how can I get the plugin id? Package::metadata() is deprecated
+                        if (!package.isValid() || !package.metadata().isValid()) {
+                            fail(i18n("The package you just dropped is invalid."));
+                            return;
+                        }
+
+                        createApplet(package.metadata().pluginId(), QVariantList(), posi);
+                    });
+                } else if (plugin.isEmpty()) {
                     //set wallpapery stuff
                     plugin = actionsToWallpapers.value(choice);
                     if (m_wallpaperInterface && tjob->url().isValid()) {
@@ -539,7 +737,7 @@ void ContainmentInterface::mimeTypeRetrieved(KIO::Job *job, const QString &mimet
             }
         } else {
             // we can at least create an icon as a link to the URL
-            Plasma::Applet *applet = createApplet("org.kde.plasma.icon", QVariantList(), posi);
+            Plasma::Applet *applet = createApplet(QStringLiteral("org.kde.plasma.icon"), QVariantList(), posi);
             setAppletArgs(applet, mimetype, tjob->url().toString());
         }
     }
@@ -554,8 +752,8 @@ void ContainmentInterface::appletAddedForward(Plasma::Applet *applet)
         return;
     }
 
-    QObject *appletGraphicObject = applet->property("_plasma_graphicObject").value<QObject *>();
-    QObject *contGraphicObject = m_containment->property("_plasma_graphicObject").value<QObject *>();
+    AppletInterface *appletGraphicObject = applet->property("_plasma_graphicObject").value<AppletInterface *>();
+    AppletInterface *contGraphicObject = m_containment->property("_plasma_graphicObject").value<AppletInterface *>();
 
 //     qDebug() << "Applet added on containment:" << m_containment->title() << contGraphicObject
 //              << "Applet: " << applet << applet->title() << appletGraphicObject;
@@ -572,14 +770,19 @@ void ContainmentInterface::appletAddedForward(Plasma::Applet *applet)
     }
 
     m_appletInterfaces << appletGraphicObject;
-    emit appletAdded(appletGraphicObject, -1, -1);
+    connect(appletGraphicObject, &QObject::destroyed, this,
+            [this](QObject *obj) {
+                m_appletInterfaces.removeAll(obj);
+            });
+    emit appletAdded(appletGraphicObject, appletGraphicObject->m_positionBeforeRemoval.x(), appletGraphicObject->m_positionBeforeRemoval.y());
     emit appletsChanged();
 }
 
 void ContainmentInterface::appletRemovedForward(Plasma::Applet *applet)
 {
-    QObject *appletGraphicObject = applet->property("_plasma_graphicObject").value<QObject *>();
+    AppletInterface *appletGraphicObject = applet->property("_plasma_graphicObject").value<AppletInterface *>();
     m_appletInterfaces.removeAll(appletGraphicObject);
+    appletGraphicObject->m_positionBeforeRemoval = appletGraphicObject->mapToItem(this, QPointF());
     emit appletRemoved(appletGraphicObject);
     emit appletsChanged();
 }
@@ -600,8 +803,8 @@ void ContainmentInterface::loadWallpaper()
         m_wallpaperInterface->setProperty("parent", QVariant::fromValue(this));
 
         //set anchors
-        QQmlExpression expr(qmlObject()->engine()->rootContext(), m_wallpaperInterface, "parent");
-        QQmlProperty prop(m_wallpaperInterface, "anchors.fill");
+        QQmlExpression expr(qmlObject()->engine()->rootContext(), m_wallpaperInterface, QStringLiteral("parent"));
+        QQmlProperty prop(m_wallpaperInterface, QStringLiteral("anchors.fill"));
         prop.write(expr.evaluate());
 
         m_containment->setProperty("wallpaperGraphicsObject", QVariant::fromValue(m_wallpaperInterface));
@@ -631,7 +834,7 @@ QList<QObject *> ContainmentInterface::actions() const
     //FIXME: giving directly a QList<QAction*> crashes
 
     QStringList actionOrder;
-    actionOrder << "add widgets" << "manage activities" << "remove" << "lock widgets" << "run associated application" << "configure";
+    actionOrder << QStringLiteral("add widgets") << QStringLiteral("manage activities") << QStringLiteral("remove") << QStringLiteral("lock widgets") << QStringLiteral("run associated application") << QStringLiteral("configure");
     QHash<QString, QAction *> orderedActions;
     //use a multimap to sort by action type
     QMultiMap<int, QObject *> actions;
@@ -670,7 +873,7 @@ QList<QObject *> ContainmentInterface::actions() const
 
     foreach (const QString &name, actionOrder) {
         QAction *a = orderedActions.value(name);
-        if (a && a->isVisible() && !a->menu()) {
+        if (a && !a->menu()) {
             actionList << a;
         }
         ++i;
@@ -708,7 +911,7 @@ void ContainmentInterface::mousePressEvent(QMouseEvent *event)
     //Don't have an action list? execute as single action
     //and set the event position as action data
     if (plugin->contextualActions().length() == 1) {
-        QAction *action = plugin->contextualActions().first();
+        QAction *action = plugin->contextualActions().at(0);
         action->setData(event->pos());
         action->trigger();
         event->accept();
@@ -719,9 +922,8 @@ void ContainmentInterface::mousePressEvent(QMouseEvent *event)
     Plasma::Applet *applet = 0;
     foreach (QObject *appletObject, m_appletInterfaces) {
         if (AppletInterface *ai = qobject_cast<AppletInterface *>(appletObject)) {
-            if (ai->contains(ai->mapFromItem(this, event->posF()))) {
+            if (ai->isVisible() && ai->contains(ai->mapFromItem(this, event->posF()))) {
                 applet = ai->applet();
-                emit ai->contextualActionsAboutToShow();
                 break;
             } else {
                 ai = 0;
@@ -735,9 +937,10 @@ void ContainmentInterface::mousePressEvent(QMouseEvent *event)
     m_contextMenu = &desktopMenu;
 
     if (applet) {
+        emit applet->contextualActionsAboutToShow();
         addAppletActions(desktopMenu, applet, event);
     } else {
-        emit contextualActionsAboutToShow();
+        emit m_containment->contextualActionsAboutToShow();
         addContainmentActions(desktopMenu, event);
     }
 
@@ -746,7 +949,23 @@ void ContainmentInterface::mousePressEvent(QMouseEvent *event)
     desktopMenu.setAttribute(Qt::WA_TranslucentBackground);
     //end workaround
 
-    desktopMenu.exec(event->globalPos());
+    if (window() && window()->mouseGrabberItem()) {
+        window()->mouseGrabberItem()->ungrabMouse();
+    }
+
+    QPoint pos = event->globalPos();
+    if (window() && applet && m_containment->containmentType() == Plasma::Types::PanelContainment) {
+        desktopMenu.adjustSize();
+
+        if (QScreen *screen = window()->screen()) {
+            const QRect geo = screen->availableGeometry();
+
+            pos = QPoint(qBound(geo.left(), pos.x(), geo.right() + 1 - desktopMenu.width()),
+                         qBound(geo.top(), pos.y(), geo.bottom() + 1 - desktopMenu.height()));
+        }
+    }
+
+    desktopMenu.exec(pos);
     event->setAccepted(true);
 }
 
@@ -755,14 +974,22 @@ void ContainmentInterface::wheelEvent(QWheelEvent *event)
     const QString trigger = Plasma::ContainmentActions::eventToString(event);
     Plasma::ContainmentActions *plugin = m_containment->containmentActions().value(trigger);
 
-    if (plugin) {
-        if (event->delta() < 0) {
-            plugin->performNextAction();
-        } else {
-            plugin->performPreviousAction();
-        }
-    } else {
+    if (!plugin) {
         event->setAccepted(false);
+        return;
+    }
+
+    m_wheelDelta += event->delta();
+
+    // Angle delta 120 for common "one click"
+    // See: http://qt-project.org/doc/qt-5/qml-qtquick-wheelevent.html#angleDelta-prop
+    while (m_wheelDelta >= 120) {
+        m_wheelDelta -= 120;
+        plugin->performPreviousAction();
+    }
+    while (m_wheelDelta <= -120) {
+        m_wheelDelta += 120;
+        plugin->performNextAction();
     }
 }
 
@@ -775,16 +1002,16 @@ void ContainmentInterface::addAppletActions(QMenu &desktopMenu, Plasma::Applet *
     }
 
     if (!applet->failedToLaunch()) {
-        QAction *runAssociatedApplication = applet->actions()->action("run associated application");
+        QAction *runAssociatedApplication = applet->actions()->action(QStringLiteral("run associated application"));
         if (runAssociatedApplication && runAssociatedApplication->isEnabled()) {
             desktopMenu.addAction(runAssociatedApplication);
         }
 
-        QAction *configureApplet = applet->actions()->action("configure");
+        QAction *configureApplet = applet->actions()->action(QStringLiteral("configure"));
         if (configureApplet && configureApplet->isEnabled()) {
             desktopMenu.addAction(configureApplet);
         }
-        QAction *appletAlternatives = applet->actions()->action("alternatives");
+        QAction *appletAlternatives = applet->actions()->action(QStringLiteral("alternatives"));
         if (appletAlternatives && appletAlternatives->isEnabled()) {
             desktopMenu.addAction(appletAlternatives);
         }
@@ -820,7 +1047,7 @@ void ContainmentInterface::addAppletActions(QMenu &desktopMenu, Plasma::Applet *
 
     if (m_containment->immutability() == Plasma::Types::Mutable &&
         (m_containment->containmentType() != Plasma::Types::PanelContainment || m_containment->isUserConfiguring())) {
-        QAction *closeApplet = applet->actions()->action("remove");
+        QAction *closeApplet = applet->actions()->action(QStringLiteral("remove"));
         //qDebug() << "checking for removal" << closeApplet;
         if (closeApplet) {
             if (!desktopMenu.isEmpty()) {
@@ -836,7 +1063,7 @@ void ContainmentInterface::addAppletActions(QMenu &desktopMenu, Plasma::Applet *
 void ContainmentInterface::addContainmentActions(QMenu &desktopMenu, QEvent *event)
 {
     if (m_containment->corona()->immutability() != Plasma::Types::Mutable &&
-            !KAuthorized::authorizeKAction("plasma/containment_actions")) {
+            !KAuthorized::authorizeKAction(QStringLiteral("plasma/containment_actions"))) {
         //qDebug() << "immutability";
         return;
     }
@@ -866,8 +1093,8 @@ void ContainmentInterface::addContainmentActions(QMenu &desktopMenu, QEvent *eve
         //a better plugin.  note that if the user sets no-plugin this won't happen...
         if ((m_containment->containmentType() != Plasma::Types::PanelContainment &&
                 m_containment->containmentType() != Plasma::Types::CustomPanelContainment) &&
-                m_containment->actions()->action("configure")) {
-            desktopMenu.addAction(m_containment->actions()->action("configure"));
+                m_containment->actions()->action(QStringLiteral("configure"))) {
+            desktopMenu.addAction(m_containment->actions()->action(QStringLiteral("configure")));
         }
     } else {
         desktopMenu.addActions(actions);
